import json
import logging
from io import BytesIO

import pytest
from bottle import Bottle, HTTPResponse, abort, redirect
from bottle import debug as set_debug
from werkzeug.test import Client
from werkzeug.wrappers import Response

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.integrations.bottle import BottleIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.serializer import MAX_DATABAG_BREADTH


@pytest.fixture(scope="function")
def app(sentry_init):
    app = Bottle()

    @app.route("/message")
    def hi():
        capture_message("hi")
        return "ok"

    @app.route("/message/<message_id>")
    def hi_with_id(message_id):
        capture_message("hi")
        return "ok"

    @app.route("/message-named-route", name="hi")
    def named_hi():
        capture_message("hi")
        return "ok"

    yield app


@pytest.fixture
def get_client(app):
    def inner():
        return Client(app)

    return inner


def test_has_context(sentry_init, app, capture_events, get_client):
    sentry_init(integrations=[BottleIntegration()])
    events = capture_events()

    client = get_client()
    response = client.get("/message")
    assert response[1] == "200 OK"

    (event,) = events
    assert event["message"] == "hi"
    assert "data" not in event["request"]
    assert event["request"]["url"] == "http://localhost/message"


@pytest.mark.parametrize(
    "url,transaction_style,expected_transaction,expected_source",
    [
        ("/message", "endpoint", "hi", "component"),
        ("/message", "url", "/message", "route"),
        ("/message/123456", "url", "/message/<message_id>", "route"),
        ("/message-named-route", "endpoint", "hi", "component"),
    ],
)
def test_transaction_style(
    sentry_init,
    url,
    transaction_style,
    expected_transaction,
    expected_source,
    capture_events,
    get_client,
):
    sentry_init(integrations=[BottleIntegration(transaction_style=transaction_style)])
    events = capture_events()

    client = get_client()
    response = client.get(url)
    assert response[1] == "200 OK"

    (event,) = events
    # We use endswith() because in Python 2.7 it is "test_bottle.hi"
    # and in later Pythons "test_bottle.app.<locals>.hi"
    assert event["transaction"].endswith(expected_transaction)
    assert event["transaction_info"] == {"source": expected_source}


@pytest.mark.parametrize("debug", (True, False), ids=["debug", "nodebug"])
@pytest.mark.parametrize("catchall", (True, False), ids=["catchall", "nocatchall"])
def test_errors(
    sentry_init, capture_exceptions, capture_events, app, debug, catchall, get_client
):
    sentry_init(integrations=[BottleIntegration()])

    app.catchall = catchall
    set_debug(mode=debug)

    exceptions = capture_exceptions()
    events = capture_events()

    @app.route("/")
    def index():
        1 / 0

    client = get_client()
    try:
        client.get("/")
    except ZeroDivisionError:
        pass

    (exc,) = exceptions
    assert isinstance(exc, ZeroDivisionError)

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "bottle"
    assert event["exception"]["values"][0]["mechanism"]["handled"] is False


@pytest.mark.parametrize("max_value_length", [1024, None])
def test_large_json_request(
    sentry_init, capture_events, app, get_client, max_value_length
):
    sentry_init(
        integrations=[BottleIntegration()],
        max_request_body_size="always",
        max_value_length=max_value_length,
    )

    data = {"foo": {"bar": "a" * (1034)}}

    @app.route("/", method="POST")
    def index():
        import bottle

        assert bottle.request.json == data
        assert bottle.request.body.read() == json.dumps(data).encode("ascii")
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()
    response = client.get("/")

    response = client.post("/", content_type="application/json", data=json.dumps(data))
    assert response[1] == "200 OK"

    (event,) = events

    if max_value_length:
        assert event["_meta"]["request"]["data"]["foo"]["bar"] == {
            "": {
                "len": 1034,
                "rem": [["!limit", "x", 1021, 1024]],
            }
        }
        assert len(event["request"]["data"]["foo"]["bar"]) == 1024
    else:
        assert len(event["request"]["data"]["foo"]["bar"]) == 1034


@pytest.mark.parametrize("data", [{}, []], ids=["empty-dict", "empty-list"])
def test_empty_json_request(sentry_init, capture_events, app, data, get_client):
    sentry_init(integrations=[BottleIntegration()])

    @app.route("/", method="POST")
    def index():
        import bottle

        assert bottle.request.json == data
        assert bottle.request.body.read() == json.dumps(data).encode("ascii")
        # assert not bottle.request.forms
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()
    response = client.post("/", content_type="application/json", data=json.dumps(data))
    assert response[1] == "200 OK"

    (event,) = events
    assert event["request"]["data"] == data


@pytest.mark.parametrize("max_value_length", [1024, None])
def test_medium_formdata_request(
    sentry_init, capture_events, app, get_client, max_value_length
):
    sentry_init(
        integrations=[BottleIntegration()],
        max_request_body_size="always",
        max_value_length=max_value_length,
    )

    data = {"foo": "a" * (1034)}

    @app.route("/", method="POST")
    def index():
        import bottle

        assert bottle.request.forms["foo"] == data["foo"]
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()
    response = client.post("/", data=data)
    assert response[1] == "200 OK"

    (event,) = events

    if max_value_length:
        assert event["_meta"]["request"]["data"]["foo"] == {
            "": {
                "len": 1034,
                "rem": [["!limit", "x", 1021, 1024]],
            }
        }
        assert len(event["request"]["data"]["foo"]) == 1024
    else:
        assert len(event["request"]["data"]["foo"]) == 1034


@pytest.mark.parametrize("input_char", ["a", b"a"])
def test_too_large_raw_request(
    sentry_init, input_char, capture_events, app, get_client
):
    sentry_init(integrations=[BottleIntegration()], max_request_body_size="small")

    data = input_char * 2000

    @app.route("/", method="POST")
    def index():
        import bottle

        if isinstance(data, bytes):
            assert bottle.request.body.read() == data
        else:
            assert bottle.request.body.read() == data.encode("ascii")
        assert not bottle.request.json
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()
    response = client.post("/", data=data)
    assert response[1] == "200 OK"

    (event,) = events
    assert event["_meta"]["request"]["data"] == {"": {"rem": [["!config", "x"]]}}
    assert not event["request"]["data"]


@pytest.mark.parametrize("max_value_length", [1024, None])
def test_files_and_form(sentry_init, capture_events, app, get_client, max_value_length):
    sentry_init(
        integrations=[BottleIntegration()],
        max_request_body_size="always",
        max_value_length=max_value_length,
    )

    data = {
        "foo": "a" * (1034),
        "file": (BytesIO(b"hello"), "hello.txt"),
    }

    @app.route("/", method="POST")
    def index():
        import bottle

        assert list(bottle.request.forms) == ["foo"]
        assert list(bottle.request.files) == ["file"]
        assert not bottle.request.json
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()
    response = client.post("/", data=data)
    assert response[1] == "200 OK"

    (event,) = events
    if max_value_length:
        assert event["_meta"]["request"]["data"]["foo"] == {
            "": {
                "len": 1034,
                "rem": [["!limit", "x", 1021, 1024]],
            }
        }
        assert len(event["request"]["data"]["foo"]) == 1024
    else:
        assert len(event["request"]["data"]["foo"]) == 1034

    assert event["_meta"]["request"]["data"]["file"] == {
        "": {
            "rem": [["!raw", "x"]],
        }
    }
    assert not event["request"]["data"]["file"]


def test_json_not_truncated_if_max_request_body_size_is_always(
    sentry_init, capture_events, app, get_client
):
    sentry_init(integrations=[BottleIntegration()], max_request_body_size="always")

    data = {
        "key{}".format(i): "value{}".format(i) for i in range(MAX_DATABAG_BREADTH + 10)
    }

    @app.route("/", method="POST")
    def index():
        import bottle

        assert bottle.request.json == data
        assert bottle.request.body.read() == json.dumps(data).encode("ascii")
        capture_message("hi")
        return "ok"

    events = capture_events()

    client = get_client()

    response = client.post("/", content_type="application/json", data=json.dumps(data))
    assert response[1] == "200 OK"

    (event,) = events
    assert event["request"]["data"] == data


@pytest.mark.parametrize(
    "integrations",
    [
        [BottleIntegration()],
        [BottleIntegration(), LoggingIntegration(event_level="ERROR")],
    ],
)
def test_errors_not_reported_twice(
    sentry_init, integrations, capture_events, app, get_client
):
    sentry_init(integrations=integrations)

    app.catchall = False

    logger = logging.getLogger("bottle.app")

    @app.route("/")
    def index():
        1 / 0

    events = capture_events()

    client = get_client()

    with pytest.raises(ZeroDivisionError):
        try:
            client.get("/")
        except ZeroDivisionError as e:
            logger.exception(e)
            raise e

    assert len(events) == 1


def test_mount(app, capture_exceptions, capture_events, sentry_init, get_client):
    sentry_init(integrations=[BottleIntegration()])

    app.catchall = False

    def crashing_app(environ, start_response):
        1 / 0

    app.mount("/wsgi/", crashing_app)

    client = Client(app)

    exceptions = capture_exceptions()
    events = capture_events()

    with pytest.raises(ZeroDivisionError) as exc:
        client.get("/wsgi/")

    (error,) = exceptions

    assert error is exc.value

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "bottle"
    assert event["exception"]["values"][0]["mechanism"]["handled"] is False


def test_error_in_errorhandler(sentry_init, capture_events, app, get_client):
    sentry_init(integrations=[BottleIntegration()])

    set_debug(False)
    app.catchall = True

    @app.route("/")
    def index():
        raise ValueError()

    @app.error(500)
    def error_handler(err):
        1 / 0

    events = capture_events()

    client = get_client()

    with pytest.raises(ZeroDivisionError):
        client.get("/")

    event1, event2 = events

    (exception,) = event1["exception"]["values"]
    assert exception["type"] == "ValueError"

    exception = event2["exception"]["values"][0]
    assert exception["type"] == "ZeroDivisionError"


def test_bad_request_not_captured(sentry_init, capture_events, app, get_client):
    sentry_init(integrations=[BottleIntegration()])
    events = capture_events()

    @app.route("/")
    def index():
        abort(400, "bad request in")

    client = get_client()

    client.get("/")

    assert not events


def test_no_exception_on_redirect(sentry_init, capture_events, app, get_client):
    sentry_init(integrations=[BottleIntegration()])
    events = capture_events()

    @app.route("/")
    def index():
        redirect("/here")

    @app.route("/here")
    def here():
        return "here"

    client = get_client()

    client.get("/")

    assert not events


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    get_client,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[BottleIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
    else:
        events = capture_events()

    client = get_client()
    client.get("/message")

    if span_streaming:
        sentry_sdk.flush()
        spans = [item.payload for item in items]
        segment = spans[-1]
        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.origin"] == "auto.http.bottle"
    else:
        (_, event) = events
        assert event["contexts"]["trace"]["origin"] == "auto.http.bottle"


@pytest.mark.parametrize("raise_error", [True, False])
@pytest.mark.parametrize(
    ("integration_kwargs", "status_code", "should_capture"),
    (
        ({}, None, False),
        ({}, 400, False),
        ({}, 451, False),  # Highest 4xx status code
        ({}, 500, True),
        ({}, 511, True),  # Highest 5xx status code
        ({"failed_request_status_codes": set()}, 500, False),
        ({"failed_request_status_codes": set()}, 511, False),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, 404, True),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, 500, True),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, 400, False),
    ),
)
def test_failed_request_status_codes(
    sentry_init,
    capture_events,
    integration_kwargs,
    status_code,
    should_capture,
    raise_error,
):
    sentry_init(integrations=[BottleIntegration(**integration_kwargs)])
    events = capture_events()

    app = Bottle()

    @app.route("/")
    def handle():
        if status_code is not None:
            response = HTTPResponse(status=status_code)
            if raise_error:
                raise response
            else:
                return response
        return "OK"

    client = Client(app, Response)
    response = client.get("/")

    expected_status = 200 if status_code is None else status_code
    assert response.status_code == expected_status

    if should_capture:
        (event,) = events
        assert event["exception"]["values"][0]["type"] == "HTTPResponse"
    else:
        assert not events


def test_failed_request_status_codes_non_http_exception(sentry_init, capture_events):
    """
    If an exception, which is not an instance of HTTPResponse, is raised, it should be captured, even if
    failed_request_status_codes is empty.
    """
    sentry_init(integrations=[BottleIntegration(failed_request_status_codes=set())])
    events = capture_events()

    app = Bottle()

    @app.route("/")
    def handle():
        1 / 0

    client = Client(app, Response)

    try:
        client.get("/")
    except ZeroDivisionError:
        pass

    (event,) = events
    assert event["exception"]["values"][0]["type"] == "ZeroDivisionError"


def test_span_streaming_basic(sentry_init, capture_items):
    sentry_init(
        integrations=[BottleIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = Bottle()

    @app.route("/message")
    def hi():
        return "ok"

    client = Client(app)
    client.get("/message")

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]

    # Segment span (root, created by WSGI middleware)
    assert segment["is_segment"] is True
    assert "parent_span_id" not in segment
    assert segment["status"] == "ok"
    assert segment["attributes"]["sentry.op"] == "http.server"
    assert segment["attributes"]["sentry.origin"] == "auto.http.bottle"
    assert segment["attributes"]["http.request.method"] == "GET"
    assert segment["attributes"]["http.response.status_code"] == 200
    assert segment["name"].endswith("hi")


@pytest.mark.parametrize(
    "url,transaction_style,expected_name,expected_source",
    [
        ("/message", "endpoint", "hi", "component"),
        ("/message", "url", "/message", "route"),
        ("/message/123456", "url", "/message/<message_id>", "route"),
        ("/message-named-route", "endpoint", "hi", "component"),
    ],
)
def test_span_streaming_transaction_style(
    sentry_init,
    capture_items,
    url,
    transaction_style,
    expected_name,
    expected_source,
):
    sentry_init(
        integrations=[BottleIntegration(transaction_style=transaction_style)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = Bottle()

    @app.route("/message")
    def hi():
        return "ok"

    @app.route("/message/<message_id>")
    def hi_with_id(message_id):
        return "ok"

    @app.route("/message-named-route", name="hi")
    def named_hi():
        return "ok"

    client = Client(app)
    client.get(url)

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]

    assert segment["is_segment"] is True

    assert segment["name"].endswith(expected_name)
    assert segment["attributes"]["sentry.segment.name.source"] == expected_source


def test_span_streaming_with_error(sentry_init, capture_items):
    sentry_init(
        integrations=[BottleIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("event", "span")

    app = Bottle()

    @app.route("/error")
    def error():
        1 / 0

    client = Client(app)
    try:
        client.get("/error")
    except ZeroDivisionError:
        pass

    sentry_sdk.flush()

    events = [item.payload for item in items if item.type == "event"]
    spans = [item.payload for item in items if item.type == "span"]
    assert len(events) == 1
    assert len(spans) == 1

    error_event = events[0]
    segment = spans[0]

    # Confirm the same trace is shared
    assert segment["trace_id"] == error_event["contexts"]["trace"]["trace_id"]

    # Span hierarchy
    assert segment["is_segment"] is True
    assert "parent_span_id" not in segment

    # Error event span_id points to the segment span (where the exception was raised)
    assert error_event["contexts"]["trace"]["span_id"] == segment["span_id"]

    # Span status
    assert segment["status"] == "error"

    # Bottle mechanism on the error event
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "bottle"
    assert error_event["exception"]["values"][0]["mechanism"]["handled"] is False


@pytest.mark.parametrize(
    "status_code,expected_span_status",
    [
        (200, "ok"),
        (404, "error"),
        (500, "error"),
    ],
)
def test_span_streaming_http_error_status(
    sentry_init,
    capture_items,
    status_code,
    expected_span_status,
):
    sentry_init(
        integrations=[BottleIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    app = Bottle()

    @app.route("/")
    def handle():
        return HTTPResponse(status=status_code, body="response")

    client = Client(app)
    client.get("/")

    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 1

    segment = spans[0]

    assert segment["is_segment"] is True

    assert segment["status"] == expected_span_status
    assert segment["attributes"]["http.response.status_code"] == status_code


@pytest.mark.parametrize("raise_error", [True, False])
@pytest.mark.parametrize(
    ("integration_kwargs", "status_code", "should_capture"),
    (
        ({}, 500, True),
        ({}, 400, False),
        ({"failed_request_status_codes": set()}, 500, False),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, 404, True),
        ({"failed_request_status_codes": {404, *range(500, 600)}}, 400, False),
    ),
)
def test_span_streaming_failed_request_status_codes(
    sentry_init,
    capture_items,
    integration_kwargs,
    status_code,
    should_capture,
    raise_error,
):
    sentry_init(
        integrations=[BottleIntegration(**integration_kwargs)],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("event", "span")

    app = Bottle()

    @app.route("/")
    def handle():
        response = HTTPResponse(status=status_code)
        if raise_error:
            raise response
        return response

    client = Client(app, Response)
    client.get("/")

    sentry_sdk.flush()

    events = [item.payload for item in items if item.type == "event"]
    spans = [item.payload for item in items if item.type == "span"]
    assert len(spans) == 1

    segment = spans[0]

    assert segment["is_segment"] is True

    if should_capture:
        assert len(events) == 1
        assert events[0]["exception"]["values"][0]["type"] == "HTTPResponse"
        assert events[0]["exception"]["values"][0]["mechanism"]["handled"] is True
    else:
        assert len(events) == 0



import json
import sys
import time
from collections.abc import Iterable
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk import consts
from sentry_sdk.ai._openai_completions_api import (
    _get_system_instructions as _get_system_instructions_completions,
)
from sentry_sdk.ai._openai_completions_api import (
    _get_text_items,
    _transform_system_instructions,
)
from sentry_sdk.ai._openai_completions_api import (
    _is_system_instruction as _is_system_instruction_completions,
)
from sentry_sdk.ai._openai_responses_api import (
    _get_system_instructions as _get_system_instructions_responses,
)
from sentry_sdk.ai._openai_responses_api import (
    _is_system_instruction as _is_system_instruction_responses,
)
from sentry_sdk.ai.monitoring import record_token_usage
from sentry_sdk.ai.utils import (
    get_start_span_function,
    normalize_message_roles,
    set_data_normalized,
    truncate_and_annotate_embedding_inputs,
    truncate_and_annotate_messages,
)
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import (
    has_span_streaming_enabled,
    should_truncate_gen_ai_input,
)
from sentry_sdk.utils import (
    capture_internal_exceptions,
    event_from_exception,
    reraise,
    safe_serialize,
)

if TYPE_CHECKING:
    from typing import (
        Any,
        AsyncIterator,
        Callable,
        Iterable,
        Iterator,
        List,
        Optional,
        Union,
    )

    from openai import Omit
    from openai.types import CompletionUsage
    from openai.types.responses import (
        ResponseInputParam,
        ResponseStreamEvent,
        SequenceNotStr,
    )
    from openai.types.responses.response_usage import ResponseUsage

    from sentry_sdk._types import TextPart
    from sentry_sdk.tracing import Span

try:
    try:
        from openai import NotGiven
    except ImportError:
        NotGiven = None

    try:
        from openai import Omit
    except ImportError:
        Omit = None

    from openai import AsyncStream, Stream
    from openai.resources import AsyncEmbeddings, Embeddings
    from openai.resources.chat.completions import AsyncCompletions, Completions

    if TYPE_CHECKING:
        from openai.types.chat import (
            ChatCompletionChunk,
            ChatCompletionMessageParam,
        )
except ImportError:
    raise DidNotEnable("OpenAI not installed")

RESPONSES_API_ENABLED = True
try:
    # responses API support was introduced in v1.66.0
    from openai.resources.responses import AsyncResponses, Responses
    from openai.types.responses.response_completed_event import ResponseCompletedEvent
except ImportError:
    RESPONSES_API_ENABLED = False


class OpenAIIntegration(Integration):
    identifier = "openai"
    origin = f"auto.ai.{identifier}"

    def __init__(
        self: "OpenAIIntegration",
        include_prompts: bool = True,
        tiktoken_encoding_name: "Optional[str]" = None,
    ) -> None:
        self.include_prompts = include_prompts

        self.tiktoken_encoding = None
        if tiktoken_encoding_name is not None:
            import tiktoken  # type: ignore

            self.tiktoken_encoding = tiktoken.get_encoding(tiktoken_encoding_name)

    @staticmethod
    def setup_once() -> None:
        Completions.create = _wrap_chat_completion_create(Completions.create)
        AsyncCompletions.create = _wrap_async_chat_completion_create(
            AsyncCompletions.create
        )

        Embeddings.create = _wrap_embeddings_create(Embeddings.create)
        AsyncEmbeddings.create = _wrap_async_embeddings_create(AsyncEmbeddings.create)

        if RESPONSES_API_ENABLED:
            Responses.create = _wrap_responses_create(Responses.create)
            AsyncResponses.create = _wrap_async_responses_create(AsyncResponses.create)

    def count_tokens(self: "OpenAIIntegration", s: str) -> int:
        if self.tiktoken_encoding is None:
            return 0
        try:
            return len(self.tiktoken_encoding.encode_ordinary(s))
        except Exception:
            return 0


def _capture_exception(exc: "Any") -> None:
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "openai", "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)


def _has_attr_and_is_int(
    token_usage: "Union[CompletionUsage, ResponseUsage]", attr_name: str
) -> bool:
    return hasattr(token_usage, attr_name) and isinstance(
        getattr(token_usage, attr_name, None), int
    )


def _calculate_completions_token_usage(
    messages: "Optional[Iterable[ChatCompletionMessageParam]]",
    response: "Any",
    span: "Union[Span, StreamedSpan]",
    streaming_message_responses: "Optional[List[str]]",
    streaming_message_total_token_usage: "Optional[CompletionUsage]",
    count_tokens: "Callable[..., Any]",
) -> None:
    """Extract and record token usage from a Chat Completions API response."""
    input_tokens: "Optional[int]" = 0
    input_tokens_cached: "Optional[int]" = 0
    output_tokens: "Optional[int]" = 0
    output_tokens_reasoning: "Optional[int]" = 0
    total_tokens: "Optional[int]" = 0
    usage = None

    if streaming_message_total_token_usage is not None:
        usage = streaming_message_total_token_usage
    elif hasattr(response, "usage"):
        usage = response.usage

    if usage is not None:
        if _has_attr_and_is_int(usage, "prompt_tokens"):
            input_tokens = usage.prompt_tokens
        if _has_attr_and_is_int(usage, "completion_tokens"):
            output_tokens = usage.completion_tokens
        if _has_attr_and_is_int(usage, "total_tokens"):
            total_tokens = usage.total_tokens

        if hasattr(usage, "prompt_tokens_details"):
            cached = getattr(usage.prompt_tokens_details, "cached_tokens", None)
            if isinstance(cached, int):
                input_tokens_cached = cached

        if hasattr(usage, "completion_tokens_details"):
            reasoning = getattr(
                usage.completion_tokens_details, "reasoning_tokens", None
            )
            if isinstance(reasoning, int):
                output_tokens_reasoning = reasoning

    # Manually count input tokens
    if input_tokens == 0:
        for message in messages or []:
            if isinstance(message, str):
                input_tokens += count_tokens(message)
                continue
            elif isinstance(message, dict):
                message_content = message.get("content")
                if message_content is None:
                    continue
                text_items = _get_text_items(message_content)
                input_tokens += sum(count_tokens(text) for text in text_items)
                continue

    # Manually count output tokens
    if output_tokens == 0:
        if streaming_message_responses is not None:
            for message in streaming_message_responses:
                output_tokens += count_tokens(message)
        elif hasattr(response, "choices") and response.choices is not None:
            for choice in response.choices:
                if hasattr(choice, "message") and hasattr(choice.message, "content"):
                    output_tokens += count_tokens(choice.message.content)

    # Do not set token data if it is 0
    input_tokens = input_tokens or None
    input_tokens_cached = input_tokens_cached or None
    output_tokens = output_tokens or None
    output_tokens_reasoning = output_tokens_reasoning or None
    total_tokens = total_tokens or None

    record_token_usage(
        span,
        input_tokens=input_tokens,
        input_tokens_cached=input_tokens_cached,
        output_tokens=output_tokens,
        output_tokens_reasoning=output_tokens_reasoning,
        total_tokens=total_tokens,
    )


def _calculate_responses_token_usage(
    input: "Any",
    response: "Any",
    span: "Union[Span, StreamedSpan]",
    streaming_message_responses: "Optional[List[str]]",
    count_tokens: "Callable[..., Any]",
) -> None:
    """Extract and record token usage from a Responses API response."""
    input_tokens: "Optional[int]" = 0
    input_tokens_cached: "Optional[int]" = 0
    output_tokens: "Optional[int]" = 0
    output_tokens_reasoning: "Optional[int]" = 0
    total_tokens: "Optional[int]" = 0

    if hasattr(response, "usage"):
        usage = response.usage

        if _has_attr_and_is_int(usage, "input_tokens"):
            input_tokens = usage.input_tokens
        if _has_attr_and_is_int(usage, "output_tokens"):
            output_tokens = usage.output_tokens
        if _has_attr_and_is_int(usage, "total_tokens"):
            total_tokens = usage.total_tokens

        if hasattr(usage, "input_tokens_details"):
            cached = getattr(usage.input_tokens_details, "cached_tokens", None)
            if isinstance(cached, int):
                input_tokens_cached = cached

        if hasattr(usage, "output_tokens_details"):
            reasoning = getattr(usage.output_tokens_details, "reasoning_tokens", None)
            if isinstance(reasoning, int):
                output_tokens_reasoning = reasoning

    # Manually count input tokens
    if input_tokens == 0:
        for message in input or []:
            if isinstance(message, str):
                input_tokens += count_tokens(message)
                continue
            elif isinstance(message, dict):
                message_content = message.get("content")
                if message_content is None:
                    continue
                # Deliberate use of Completions function for both Completions and Responses input format.
                text_items = _get_text_items(message_content)
                input_tokens += sum(count_tokens(text) for text in text_items)
                continue

    # Manually count output tokens
    if output_tokens == 0:
        if streaming_message_responses is not None:
            for message in streaming_message_responses:
                output_tokens += count_tokens(message)
        elif hasattr(response, "output"):
            for output_item in response.output:
                if hasattr(output_item, "content"):
                    for content_item in output_item.content:
                        if hasattr(content_item, "text"):
                            output_tokens += count_tokens(content_item.text)

    # Do not set token data if it is 0
    input_tokens = input_tokens or None
    input_tokens_cached = input_tokens_cached or None
    output_tokens = output_tokens or None
    output_tokens_reasoning = output_tokens_reasoning or None
    total_tokens = total_tokens or None

    record_token_usage(
        span,
        input_tokens=input_tokens,
        input_tokens_cached=input_tokens_cached,
        output_tokens=output_tokens,
        output_tokens_reasoning=output_tokens_reasoning,
        total_tokens=total_tokens,
    )


def _set_responses_api_input_data(
    span: "Union[Span, StreamedSpan]",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
) -> None:
    explicit_instructions: "Union[Optional[str], Omit]" = kwargs.get("instructions")
    messages: "Optional[Union[str, ResponseInputParam]]" = kwargs.get("input")

    tools = kwargs.get("tools")
    if tools is not None and _is_given(tools) and len(tools) > 0:
        set_data_normalized(
            span, SPANDATA.GEN_AI_REQUEST_AVAILABLE_TOOLS, safe_serialize(tools)
        )

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    model = kwargs.get("model")
    if model is not None:
        set_on_span(SPANDATA.GEN_AI_REQUEST_MODEL, model)

    max_tokens = kwargs.get("max_output_tokens")
    if max_tokens is not None and _is_given(max_tokens):
        set_on_span(SPANDATA.GEN_AI_REQUEST_MAX_TOKENS, max_tokens)

    temperature = kwargs.get("temperature")
    if temperature is not None and _is_given(temperature):
        set_on_span(SPANDATA.GEN_AI_REQUEST_TEMPERATURE, temperature)

    top_p = kwargs.get("top_p")
    if top_p is not None and _is_given(top_p):
        set_on_span(SPANDATA.GEN_AI_REQUEST_TOP_P, top_p)

    conversation = kwargs.get("conversation")
    if conversation is not None and _is_given(conversation):
        conversation_id: "Optional[str]" = None
        if isinstance(conversation, str):
            conversation_id = conversation
        elif isinstance(conversation, dict):
            conversation_id = conversation.get("id")
        if conversation_id is not None:
            set_on_span(SPANDATA.GEN_AI_CONVERSATION_ID, conversation_id)

    if not should_send_default_pii() or not integration.include_prompts:
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "responses")
        return

    if (
        messages is None
        and explicit_instructions is not None
        and _is_given(explicit_instructions)
    ):
        set_on_span(
            SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
            json.dumps(
                [
                    {
                        "type": "text",
                        "content": explicit_instructions,
                    }
                ]
            ),
        )

        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "responses")
        return

    if messages is None:
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "responses")
        return

    instructions_text_parts: "list[TextPart]" = []
    if explicit_instructions is not None and _is_given(explicit_instructions):
        instructions_text_parts.append(
            {
                "type": "text",
                "content": explicit_instructions,
            }
        )

    system_instructions = _get_system_instructions_responses(messages)
    # Deliberate use of function accepting completions API type because
    # of shared structure FOR THIS PURPOSE ONLY.
    instructions_text_parts += _transform_system_instructions(system_instructions)

    if len(instructions_text_parts) > 0:
        set_on_span(
            SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
            json.dumps(instructions_text_parts),
        )

    if isinstance(messages, str):
        normalized_messages = normalize_message_roles([messages])  # type: ignore
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_messages(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_REQUEST_MESSAGES, messages_data, unpack=False
            )

        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "responses")
        return

    non_system_messages = [
        message for message in messages if not _is_system_instruction_responses(message)
    ]
    if len(non_system_messages) > 0:
        normalized_messages = normalize_message_roles(non_system_messages)
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_messages(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_REQUEST_MESSAGES, messages_data, unpack=False
            )

    set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "responses")


def _set_completions_api_input_data(
    span: "Union[Span, StreamedSpan]",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
) -> None:
    messages: "Optional[Union[str, Iterable[ChatCompletionMessageParam]]]" = kwargs.get(
        "messages"
    )

    tools = kwargs.get("tools")
    if tools is not None and _is_given(tools) and len(tools) > 0:
        set_data_normalized(
            span, SPANDATA.GEN_AI_REQUEST_AVAILABLE_TOOLS, safe_serialize(tools)
        )

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    model = kwargs.get("model")
    if model is not None:
        set_on_span(SPANDATA.GEN_AI_REQUEST_MODEL, model)

    max_tokens = kwargs.get("max_tokens")
    if max_tokens is not None and _is_given(max_tokens):
        set_on_span(SPANDATA.GEN_AI_REQUEST_MAX_TOKENS, max_tokens)

    presence_penalty = kwargs.get("presence_penalty")
    if presence_penalty is not None and _is_given(presence_penalty):
        set_on_span(SPANDATA.GEN_AI_REQUEST_PRESENCE_PENALTY, presence_penalty)

    frequency_penalty = kwargs.get("frequency_penalty")
    if frequency_penalty is not None and _is_given(frequency_penalty):
        set_on_span(SPANDATA.GEN_AI_REQUEST_FREQUENCY_PENALTY, frequency_penalty)

    temperature = kwargs.get("temperature")
    if temperature is not None and _is_given(temperature):
        set_on_span(SPANDATA.GEN_AI_REQUEST_TEMPERATURE, temperature)

    top_p = kwargs.get("top_p")
    if top_p is not None and _is_given(top_p):
        set_on_span(SPANDATA.GEN_AI_REQUEST_TOP_P, top_p)

    if (
        not should_send_default_pii()
        or not integration.include_prompts
        or messages is None
    ):
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "chat")
        return

    if isinstance(messages, str):
        normalized_messages = normalize_message_roles([messages])  # type: ignore
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_messages(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_REQUEST_MESSAGES, messages_data, unpack=False
            )
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "chat")
        return

    # dict special case following https://github.com/openai/openai-python/blob/3e0c05b84a2056870abf3bd6a5e7849020209cc3/src/openai/_utils/_transform.py#L194-L197
    if not isinstance(messages, Iterable) or isinstance(messages, dict):
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "chat")
        return

    messages = list(messages)
    kwargs["messages"] = messages

    system_instructions = _get_system_instructions_completions(messages)
    if len(system_instructions) > 0:
        set_on_span(
            SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
            json.dumps(_transform_system_instructions(system_instructions)),
        )

    non_system_messages = [
        message
        for message in messages
        if not _is_system_instruction_completions(message)
    ]
    if len(non_system_messages) > 0:
        normalized_messages = normalize_message_roles(non_system_messages)
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_messages(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_REQUEST_MESSAGES, messages_data, unpack=False
            )

    set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "chat")


def _set_embeddings_input_data(
    span: "Union[Span, StreamedSpan]",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
) -> None:
    messages: "Union[str, SequenceNotStr[str], Iterable[int], Iterable[Iterable[int]]]" = kwargs.get(
        "input"
    )

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    model = kwargs.get("model")
    if model is not None:
        set_on_span(SPANDATA.GEN_AI_REQUEST_MODEL, model)

    if (
        not should_send_default_pii()
        or not integration.include_prompts
        or messages is None
    ):
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")

        return

    if isinstance(messages, str):
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")

        normalized_messages = normalize_message_roles([messages])  # type: ignore
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_embedding_inputs(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_EMBEDDINGS_INPUT, messages_data, unpack=False
            )

        return

    # dict special case following https://github.com/openai/openai-python/blob/3e0c05b84a2056870abf3bd6a5e7849020209cc3/src/openai/_utils/_transform.py#L194-L197
    if not isinstance(messages, Iterable) or isinstance(messages, dict):
        set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")
        return

    messages = list(messages)
    kwargs["input"] = messages

    if len(messages) > 0:
        normalized_messages = normalize_message_roles(messages)
        client = sentry_sdk.get_client()
        scope = sentry_sdk.get_current_scope()
        messages_data = (
            truncate_and_annotate_embedding_inputs(normalized_messages, span, scope)
            if should_truncate_gen_ai_input(client.options)
            else normalized_messages
        )
        if messages_data is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_EMBEDDINGS_INPUT, messages_data, unpack=False
            )

    set_data_normalized(span, SPANDATA.GEN_AI_OPERATION_NAME, "embeddings")


def _set_common_output_data(
    span: "Union[Span, StreamedSpan]",
    response: "Any",
    input: "Any",
    integration: "OpenAIIntegration",
    finish_span: bool = True,
) -> None:
    if hasattr(response, "model"):
        set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_MODEL, response.model)

    # Chat Completions API
    if hasattr(response, "choices") and response.choices is not None:
        if should_send_default_pii() and integration.include_prompts:
            response_text = [
                choice.message.model_dump()
                for choice in response.choices
                if choice.message is not None
            ]
            if len(response_text) > 0:
                set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, response_text)

        _calculate_completions_token_usage(
            messages=input,
            response=response,
            span=span,
            streaming_message_responses=None,
            streaming_message_total_token_usage=None,
            count_tokens=integration.count_tokens,
        )

        if finish_span:
            span.__exit__(None, None, None)

    # Responses API
    elif hasattr(response, "output"):
        if should_send_default_pii() and integration.include_prompts:
            output_messages: "dict[str, list[Any]]" = {
                "response": [],
                "tool": [],
            }

            for output in response.output:
                if output.type == "function_call":
                    output_messages["tool"].append(output.dict())
                elif output.type == "message":
                    for output_message in output.content:
                        try:
                            output_messages["response"].append(output_message.text)
                        except AttributeError:
                            # Unknown output message type, just return the json
                            output_messages["response"].append(output_message.dict())

            if len(output_messages["tool"]) > 0:
                set_data_normalized(
                    span,
                    SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS,
                    output_messages["tool"],
                    unpack=False,
                )

            if len(output_messages["response"]) > 0:
                set_data_normalized(
                    span, SPANDATA.GEN_AI_RESPONSE_TEXT, output_messages["response"]
                )

        _calculate_responses_token_usage(
            input=input,
            response=response,
            span=span,
            streaming_message_responses=None,
            count_tokens=integration.count_tokens,
        )

        if finish_span:
            span.__exit__(None, None, None)
    # Embeddings API (fallback for responses with neither choices nor output)
    else:
        _calculate_completions_token_usage(
            messages=input,
            response=response,
            span=span,
            streaming_message_responses=None,
            streaming_message_total_token_usage=None,
            count_tokens=integration.count_tokens,
        )
        if finish_span:
            span.__exit__(None, None, None)


def _new_sync_chat_completion(f: "Any", *args: "Any", **kwargs: "Any") -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return f(*args, **kwargs)

    if "messages" not in kwargs:
        # invalid call (in all versions of openai), let it return error
        return f(*args, **kwargs)

    try:
        iter(kwargs["messages"])
    except TypeError:
        # invalid call (in all versions), messages must be iterable
        return f(*args, **kwargs)

    model = kwargs.get("model")

    # Same bool handling as in https://github.com/openai/openai-python/blob/acd0c54d8a68efeedde0e5b4e6c310eef1ce7867/src/openai/resources/completions.py#L585
    is_streaming_response = kwargs.get("stream", False) or False

    if has_span_streaming_enabled(client.options):
        span = sentry_sdk.traces.start_span(
            name=f"chat {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_CHAT,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
                SPANDATA.GEN_AI_RESPONSE_STREAMING: is_streaming_response,
            },
        )

    else:
        span = get_start_span_function()(
            op=consts.OP.GEN_AI_CHAT,
            name=f"chat {model}",
            origin=OpenAIIntegration.origin,
        )
        span.__enter__()

        span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
        span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, is_streaming_response)

    _set_completions_api_input_data(span, kwargs, integration)

    start_time = time.perf_counter()

    try:
        response = f(*args, **kwargs)
    except Exception as exc:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            _capture_exception(exc)
            span.__exit__(*exc_info)
        reraise(*exc_info)

    # Attribute check to fail gracefully if the attribute is not present in future `openai` versions.
    if isinstance(response, Stream) and hasattr(response, "_iterator"):
        messages = kwargs.get("messages")

        if messages is not None and isinstance(messages, str):
            messages = [messages]

        response._iterator = _wrap_synchronous_completions_chunk_iterator(
            span=span,
            integration=integration,
            start_time=start_time,
            messages=messages,
            response=response,
            old_iterator=response._iterator,
            finish_span=True,
        )

    else:
        _set_completions_api_output_data(
            span, response, kwargs, integration, finish_span=True
        )

    return response


async def _new_async_chat_completion(f: "Any", *args: "Any", **kwargs: "Any") -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return await f(*args, **kwargs)

    if "messages" not in kwargs:
        # invalid call (in all versions of openai), let it return error
        return await f(*args, **kwargs)

    try:
        iter(kwargs["messages"])
    except TypeError:
        # invalid call (in all versions), messages must be iterable
        return await f(*args, **kwargs)

    model = kwargs.get("model")

    # Same bool handling as in https://github.com/openai/openai-python/blob/acd0c54d8a68efeedde0e5b4e6c310eef1ce7867/src/openai/resources/completions.py#L585
    is_streaming_response = kwargs.get("stream", False) or False

    if has_span_streaming_enabled(client.options):
        span = sentry_sdk.traces.start_span(
            name=f"chat {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_CHAT,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
                SPANDATA.GEN_AI_RESPONSE_STREAMING: is_streaming_response,
            },
        )
    else:
        span = get_start_span_function()(
            op=consts.OP.GEN_AI_CHAT,
            name=f"chat {model}",
            origin=OpenAIIntegration.origin,
        )
        span.__enter__()

        span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
        span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, is_streaming_response)

    _set_completions_api_input_data(span, kwargs, integration)

    start_time = time.perf_counter()

    try:
        response = await f(*args, **kwargs)
    except Exception as exc:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            _capture_exception(exc)
            span.__exit__(*exc_info)
        reraise(*exc_info)

    # Attribute check to fail gracefully if the attribute is not present in future `openai` versions.
    if isinstance(response, AsyncStream) and hasattr(response, "_iterator"):
        messages = kwargs.get("messages")

        if messages is not None and isinstance(messages, str):
            messages = [messages]

        response._iterator = _wrap_asynchronous_completions_chunk_iterator(
            span=span,
            integration=integration,
            start_time=start_time,
            messages=messages,
            response=response,
            old_iterator=response._iterator,
            finish_span=True,
        )
    else:
        _set_completions_api_output_data(
            span, response, kwargs, integration, finish_span=True
        )

    return response


def _set_completions_api_output_data(
    span: "Union[Span, StreamedSpan]",
    response: "Any",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
    finish_span: bool = True,
) -> None:
    messages = kwargs.get("messages")

    if messages is not None and isinstance(messages, str):
        messages = [messages]

    _set_common_output_data(
        span,
        response,
        messages,
        integration,
        finish_span,
    )


def _wrap_synchronous_completions_chunk_iterator(
    span: "Union[Span, StreamedSpan]",
    integration: "OpenAIIntegration",
    start_time: "Optional[float]",
    messages: "Optional[Iterable[ChatCompletionMessageParam]]",
    response: "Stream[ChatCompletionChunk]",
    old_iterator: "Iterator[ChatCompletionChunk]",
    finish_span: "bool",
) -> "Iterator[ChatCompletionChunk]":
    """
    Sets information received while iterating the response stream on the AI Client Span.
    Compute token count based on inputs and outputs using tiktoken if token counts are not in the model response.
    Responsible for closing the AI Client Span if instructed to by the `finish_span` argument.
    """
    ttft = None
    data_buf: "list[list[str]]" = []  # one for each choice
    streaming_message_total_token_usage = None

    for x in old_iterator:
        if isinstance(span, StreamedSpan):
            span.set_attribute(SPANDATA.GEN_AI_RESPONSE_MODEL, x.model)
        else:
            span.set_data(SPANDATA.GEN_AI_RESPONSE_MODEL, x.model)

        with capture_internal_exceptions():
            if hasattr(x, "choices") and x.choices is not None:
                choice_index = 0
                for choice in x.choices:
                    if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
                        if start_time is not None and ttft is None:
                            ttft = time.perf_counter() - start_time
                        content = choice.delta.content
                        if len(data_buf) <= choice_index:
                            data_buf.append([])
                        data_buf[choice_index].append(content or "")
                    choice_index += 1
            if hasattr(x, "usage"):
                streaming_message_total_token_usage = x.usage

        yield x

    with capture_internal_exceptions():
        if ttft is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_RESPONSE_TIME_TO_FIRST_TOKEN, ttft
            )
        all_responses = None
        if len(data_buf) > 0:
            all_responses = ["".join(chunk) for chunk in data_buf]
            if should_send_default_pii() and integration.include_prompts:
                set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, all_responses)

        _calculate_completions_token_usage(
            messages=messages,
            response=response,
            span=span,
            streaming_message_responses=all_responses,
            streaming_message_total_token_usage=streaming_message_total_token_usage,
            count_tokens=integration.count_tokens,
        )

    if finish_span:
        span.__exit__(None, None, None)


async def _wrap_asynchronous_completions_chunk_iterator(
    span: "Union[Span, StreamedSpan]",
    integration: "OpenAIIntegration",
    start_time: "Optional[float]",
    messages: "Optional[Iterable[ChatCompletionMessageParam]]",
    response: "AsyncStream[ChatCompletionChunk]",
    old_iterator: "AsyncIterator[ChatCompletionChunk]",
    finish_span: "bool",
) -> "AsyncIterator[ChatCompletionChunk]":
    """
    Sets information received while iterating the response stream on the AI Client Span.
    Compute token count based on inputs and outputs using tiktoken if token counts are not in the model response.
    Responsible for closing the AI Client Span if instructed to by the `finish_span` argument.
    """
    ttft = None
    data_buf: "list[list[str]]" = []  # one for each choice
    streaming_message_total_token_usage = None

    async for x in old_iterator:
        if isinstance(span, StreamedSpan):
            span.set_attribute(SPANDATA.GEN_AI_RESPONSE_MODEL, x.model)
        else:
            span.set_data(SPANDATA.GEN_AI_RESPONSE_MODEL, x.model)

        with capture_internal_exceptions():
            if hasattr(x, "choices") and x.choices is not None:
                choice_index = 0
                for choice in x.choices:
                    if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
                        if start_time is not None and ttft is None:
                            ttft = time.perf_counter() - start_time
                        content = choice.delta.content
                        if len(data_buf) <= choice_index:
                            data_buf.append([])
                        data_buf[choice_index].append(content or "")
                    choice_index += 1
            if hasattr(x, "usage"):
                streaming_message_total_token_usage = x.usage

        yield x

    with capture_internal_exceptions():
        if ttft is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_RESPONSE_TIME_TO_FIRST_TOKEN, ttft
            )
        all_responses = None
        if len(data_buf) > 0:
            all_responses = ["".join(chunk) for chunk in data_buf]
            if should_send_default_pii() and integration.include_prompts:
                set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, all_responses)

        _calculate_completions_token_usage(
            messages=messages,
            response=response,
            span=span,
            streaming_message_responses=all_responses,
            streaming_message_total_token_usage=streaming_message_total_token_usage,
            count_tokens=integration.count_tokens,
        )

    if finish_span:
        span.__exit__(None, None, None)


def _wrap_synchronous_responses_event_iterator(
    span: "Union[Span, StreamedSpan]",
    integration: "OpenAIIntegration",
    start_time: "Optional[float]",
    input: "Optional[Union[str, ResponseInputParam]]",
    response: "Stream[ResponseStreamEvent]",
    old_iterator: "Iterator[ResponseStreamEvent]",
    finish_span: "bool",
) -> "Iterator[ResponseStreamEvent]":
    """
    Sets information received while iterating the response stream on the AI Client Span.
    Compute token count based on inputs and outputs using tiktoken if token counts are not in the model response.
    Responsible for closing the AI Client Span if instructed to by the `finish_span` argument.
    """
    ttft = None
    data_buf: "list[list[str]]" = []  # one for each choice

    count_tokens_manually = True
    for x in old_iterator:
        with capture_internal_exceptions():
            if hasattr(x, "delta"):
                if start_time is not None and ttft is None:
                    ttft = time.perf_counter() - start_time
                if len(data_buf) == 0:
                    data_buf.append([])
                data_buf[0].append(x.delta or "")

            if isinstance(x, ResponseCompletedEvent):
                if isinstance(span, StreamedSpan):
                    span.set_attribute(SPANDATA.GEN_AI_RESPONSE_MODEL, x.response.model)
                else:
                    span.set_data(SPANDATA.GEN_AI_RESPONSE_MODEL, x.response.model)

                _calculate_responses_token_usage(
                    input=input,
                    response=x.response,
                    span=span,
                    streaming_message_responses=None,
                    count_tokens=integration.count_tokens,
                )
                count_tokens_manually = False

        yield x

    with capture_internal_exceptions():
        if ttft is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_RESPONSE_TIME_TO_FIRST_TOKEN, ttft
            )
        if len(data_buf) > 0:
            all_responses = ["".join(chunk) for chunk in data_buf]
            if should_send_default_pii() and integration.include_prompts:
                set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, all_responses)

            if count_tokens_manually:
                _calculate_responses_token_usage(
                    input=input,
                    response=response,
                    span=span,
                    streaming_message_responses=all_responses,
                    count_tokens=integration.count_tokens,
                )

    if finish_span:
        span.__exit__(None, None, None)


async def _wrap_asynchronous_responses_event_iterator(
    span: "Union[Span, StreamedSpan]",
    integration: "OpenAIIntegration",
    start_time: "Optional[float]",
    input: "Optional[Union[str, ResponseInputParam]]",
    response: "AsyncStream[ResponseStreamEvent]",
    old_iterator: "AsyncIterator[ResponseStreamEvent]",
    finish_span: "bool",
) -> "AsyncIterator[ResponseStreamEvent]":
    """
    Sets information received while iterating the response stream on the AI Client Span.
    Compute token count based on inputs and outputs using tiktoken if token counts are not in the model response.
    Responsible for closing the AI Client Span if instructed to by the `finish_span` argument.
    """
    ttft: "Optional[float]" = None
    data_buf: "list[list[str]]" = []  # one for each choice

    count_tokens_manually = True
    async for x in old_iterator:
        with capture_internal_exceptions():
            if hasattr(x, "delta"):
                if start_time is not None and ttft is None:
                    ttft = time.perf_counter() - start_time
                if len(data_buf) == 0:
                    data_buf.append([])
                data_buf[0].append(x.delta or "")

            if isinstance(x, ResponseCompletedEvent):
                if isinstance(span, StreamedSpan):
                    span.set_attribute(SPANDATA.GEN_AI_RESPONSE_MODEL, x.response.model)
                else:
                    span.set_data(SPANDATA.GEN_AI_RESPONSE_MODEL, x.response.model)

                _calculate_responses_token_usage(
                    input=input,
                    response=x.response,
                    span=span,
                    streaming_message_responses=None,
                    count_tokens=integration.count_tokens,
                )
                count_tokens_manually = False

        yield x

    with capture_internal_exceptions():
        if ttft is not None:
            set_data_normalized(
                span, SPANDATA.GEN_AI_RESPONSE_TIME_TO_FIRST_TOKEN, ttft
            )
        if len(data_buf) > 0:
            all_responses = ["".join(chunk) for chunk in data_buf]
            if should_send_default_pii() and integration.include_prompts:
                set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, all_responses)
            if count_tokens_manually:
                _calculate_responses_token_usage(
                    input=input,
                    response=response,
                    span=span,
                    streaming_message_responses=all_responses,
                    count_tokens=integration.count_tokens,
                )
    if finish_span:
        span.__exit__(None, None, None)


def _set_responses_api_output_data(
    span: "Union[Span, StreamedSpan]",
    response: "Any",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
    finish_span: bool = True,
) -> None:
    input = kwargs.get("input")

    if input is not None and isinstance(input, str):
        input = [input]

    _set_common_output_data(
        span,
        response,
        input,
        integration,
        finish_span,
    )


def _set_embeddings_output_data(
    span: "Union[Span, StreamedSpan]",
    response: "Any",
    kwargs: "dict[str, Any]",
    integration: "OpenAIIntegration",
    finish_span: bool = True,
) -> None:
    input = kwargs.get("input")

    if input is not None and isinstance(input, str):
        input = [input]

    _set_common_output_data(
        span,
        response,
        input,
        integration,
        finish_span,
    )


def _wrap_chat_completion_create(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    def _sentry_patched_create_sync(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None or "messages" not in kwargs:
            # no "messages" means invalid call (in all versions of openai), let it return error
            return f(*args, **kwargs)

        return _new_sync_chat_completion(f, *args, **kwargs)

    return _sentry_patched_create_sync


def _wrap_async_chat_completion_create(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    async def _sentry_patched_create_async(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None or "messages" not in kwargs:
            # no "messages" means invalid call (in all versions of openai), let it return error
            return await f(*args, **kwargs)

        return await _new_async_chat_completion(f, *args, **kwargs)

    return _sentry_patched_create_async


def _new_sync_embeddings_create(f: "Any", *args: "Any", **kwargs: "Any") -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return f(*args, **kwargs)

    model = kwargs.get("model")

    if has_span_streaming_enabled(client.options):
        with sentry_sdk.traces.start_span(
            name=f"embeddings {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_EMBEDDINGS,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
            },
        ) as span:
            _set_embeddings_input_data(span, kwargs, integration)

            try:
                response = f(*args, **kwargs)
            except Exception as exc:
                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    _capture_exception(exc)
                reraise(*exc_info)

            _set_embeddings_output_data(
                span, response, kwargs, integration, finish_span=False
            )

            return response
    else:
        with get_start_span_function()(
            op=consts.OP.GEN_AI_EMBEDDINGS,
            name=f"embeddings {model}",
            origin=OpenAIIntegration.origin,
        ) as span:
            span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
            _set_embeddings_input_data(span, kwargs, integration)

            try:
                response = f(*args, **kwargs)
            except Exception as exc:
                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    _capture_exception(exc)
                reraise(*exc_info)

            _set_embeddings_output_data(
                span, response, kwargs, integration, finish_span=False
            )

            return response


async def _new_async_embeddings_create(
    f: "Any", *args: "Any", **kwargs: "Any"
) -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return await f(*args, **kwargs)

    model = kwargs.get("model")

    if has_span_streaming_enabled(client.options):
        with sentry_sdk.traces.start_span(
            name=f"embeddings {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_EMBEDDINGS,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
            },
        ) as span:
            _set_embeddings_input_data(span, kwargs, integration)

            try:
                response = await f(*args, **kwargs)
            except Exception as exc:
                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    _capture_exception(exc)
                reraise(*exc_info)

            _set_embeddings_output_data(
                span, response, kwargs, integration, finish_span=False
            )

            return response
    else:
        with get_start_span_function()(
            op=consts.OP.GEN_AI_EMBEDDINGS,
            name=f"embeddings {model}",
            origin=OpenAIIntegration.origin,
        ) as span:
            span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
            _set_embeddings_input_data(span, kwargs, integration)

            try:
                response = await f(*args, **kwargs)
            except Exception as exc:
                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    _capture_exception(exc)
                reraise(*exc_info)

            _set_embeddings_output_data(
                span, response, kwargs, integration, finish_span=False
            )

            return response


def _wrap_embeddings_create(f: "Any") -> "Any":
    @wraps(f)
    def _sentry_patched_create_sync(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None:
            return f(*args, **kwargs)

        return _new_sync_embeddings_create(f, *args, **kwargs)

    return _sentry_patched_create_sync


def _wrap_async_embeddings_create(f: "Any") -> "Any":
    @wraps(f)
    async def _sentry_patched_create_async(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None:
            return await f(*args, **kwargs)

        return await _new_async_embeddings_create(f, *args, **kwargs)

    return _sentry_patched_create_async


def _new_sync_responses_create(f: "Any", *args: "Any", **kwargs: "Any") -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return f(*args, **kwargs)

    model = kwargs.get("model")

    # Same bool handling as in https://github.com/openai/openai-python/blob/acd0c54d8a68efeedde0e5b4e6c310eef1ce7867/src/openai/resources/responses/responses.py#L940
    is_streaming_response = kwargs.get("stream", False) or False

    if has_span_streaming_enabled(client.options):
        span = sentry_sdk.traces.start_span(
            name=f"responses {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_RESPONSES,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
                SPANDATA.GEN_AI_RESPONSE_STREAMING: is_streaming_response,
            },
        )
    else:
        span = get_start_span_function()(
            op=consts.OP.GEN_AI_RESPONSES,
            name=f"responses {model}",
            origin=OpenAIIntegration.origin,
        )
        span.__enter__()

        span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
        span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, is_streaming_response)

    _set_responses_api_input_data(span, kwargs, integration)

    start_time = time.perf_counter()

    try:
        response = f(*args, **kwargs)
    except Exception as exc:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            _capture_exception(exc)
            span.__exit__(*exc_info)
        reraise(*exc_info)

    # Attribute check to fail gracefully if the attribute is not present in future `openai` versions.
    if isinstance(response, Stream) and hasattr(response, "_iterator"):
        input = kwargs.get("input")

        if input is not None and isinstance(input, str):
            input = [input]

        response._iterator = _wrap_synchronous_responses_event_iterator(
            span=span,
            integration=integration,
            start_time=start_time,
            input=input,
            response=response,
            old_iterator=response._iterator,
            finish_span=True,
        )

    else:
        _set_responses_api_output_data(
            span, response, kwargs, integration, finish_span=True
        )

    return response


async def _new_async_responses_create(f: "Any", *args: "Any", **kwargs: "Any") -> "Any":
    client = sentry_sdk.get_client()
    integration = client.get_integration(OpenAIIntegration)
    if integration is None:
        return await f(*args, **kwargs)

    model = kwargs.get("model")

    # Same bool handling as in https://github.com/openai/openai-python/blob/acd0c54d8a68efeedde0e5b4e6c310eef1ce7867/src/openai/resources/responses/responses.py#L940
    is_streaming_response = kwargs.get("stream", False) or False

    if has_span_streaming_enabled(client.options):
        span = sentry_sdk.traces.start_span(
            name=f"responses {model}",
            attributes={
                "sentry.op": consts.OP.GEN_AI_RESPONSES,
                "sentry.origin": OpenAIIntegration.origin,
                SPANDATA.GEN_AI_SYSTEM: "openai",
                SPANDATA.GEN_AI_RESPONSE_STREAMING: is_streaming_response,
            },
        )
    else:
        span = get_start_span_function()(
            op=consts.OP.GEN_AI_RESPONSES,
            name=f"responses {model}",
            origin=OpenAIIntegration.origin,
        )
        span.__enter__()

        span.set_data(SPANDATA.GEN_AI_SYSTEM, "openai")
        span.set_data(SPANDATA.GEN_AI_RESPONSE_STREAMING, is_streaming_response)

    _set_responses_api_input_data(span, kwargs, integration)

    start_time = time.perf_counter()

    try:
        response = await f(*args, **kwargs)
    except Exception as exc:
        exc_info = sys.exc_info()
        with capture_internal_exceptions():
            _capture_exception(exc)
            span.__exit__(*exc_info)
        reraise(*exc_info)

    # Attribute check to fail gracefully if the attribute is not present in future `openai` versions.
    if isinstance(response, AsyncStream) and hasattr(response, "_iterator"):
        input = kwargs.get("input")

        if input is not None and isinstance(input, str):
            input = [input]

        response._iterator = _wrap_asynchronous_responses_event_iterator(
            span=span,
            integration=integration,
            start_time=start_time,
            input=input,
            response=response,
            old_iterator=response._iterator,
            finish_span=True,
        )
    else:
        _set_responses_api_output_data(
            span, response, kwargs, integration, finish_span=True
        )

    return response


def _wrap_responses_create(f: "Any") -> "Any":
    @wraps(f)
    def _sentry_patched_create_sync(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None:
            return f(*args, **kwargs)

        return _new_sync_responses_create(f, *args, **kwargs)

    return _sentry_patched_create_sync


def _wrap_async_responses_create(f: "Any") -> "Any":
    @wraps(f)
    async def _sentry_patched_responses_async(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(OpenAIIntegration)
        if integration is None:
            return await f(*args, **kwargs)

        return await _new_async_responses_create(f, *args, **kwargs)

    return _sentry_patched_responses_async


def _is_given(obj: "Any") -> bool:
    """
    Check for givenness safely across different openai versions.
    """
    if NotGiven is not None and isinstance(obj, NotGiven):
        return False
    if Omit is not None and isinstance(obj, Omit):
        return False
    return True



import os
import sys

import pytest
from celery.contrib.testing.worker import start_worker

from sentry_sdk.utils import logger
from tests.integrations.celery.integration_tests import run_beat

REDIS_SERVER = (
    f"redis://{os.environ.get('SENTRY_PYTHON_TEST_REDIS_HOST', '127.0.0.1')}:6379"
)
REDIS_DB = 15


@pytest.fixture()
def celery_config():
    return {
        "worker_concurrency": 1,
        "broker_url": f"{REDIS_SERVER}/{REDIS_DB}",
        "result_backend": f"{REDIS_SERVER}/{REDIS_DB}",
        "beat_scheduler": "tests.integrations.celery.integration_tests:ImmediateScheduler",
        "task_always_eager": False,
        "task_create_missing_queues": True,
        "task_default_queue": f"queue_{os.getpid()}",
    }


@pytest.fixture
def celery_init(sentry_init, celery_config):
    """
    Create a Sentry instrumented Celery app.
    """
    from celery import Celery

    from sentry_sdk.integrations.celery import CeleryIntegration

    def inner(propagate_traces=True, monitor_beat_tasks=False, **kwargs):
        sentry_init(
            integrations=[
                CeleryIntegration(
                    propagate_traces=propagate_traces,
                    monitor_beat_tasks=monitor_beat_tasks,
                )
            ],
            **kwargs,
        )
        app = Celery("tasks")
        app.conf.update(celery_config)

        return app

    return inner


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
@pytest.mark.forked
def test_explanation(celery_init, capture_envelopes):
    """
    This is a dummy test for explaining how to test using Celery Beat
    """

    # First initialize a Celery app.
    # You can give the options of CeleryIntegrations
    # and the options for `sentry_dks.init` as keyword arguments.
    # See the celery_init fixture for details.
    app = celery_init(
        monitor_beat_tasks=True,
    )

    # Capture envelopes.
    envelopes = capture_envelopes()

    # Define the task you want to run
    @app.task
    def test_task():
        logger.info("Running test_task")

    # Add the task to the beat schedule
    app.add_periodic_task(60.0, test_task.s(), name="success_from_beat")

    # Start a Celery worker
    with start_worker(app, perform_ping_check=False):
        # And start a Celery Beat instance
        # This Celery Beat will start the task above immediately
        # after start for the first time
        # By default Celery Beat is terminated after 1 second.
        # See `run_beat` function on how to change this.
        run_beat(app)

    # After the Celery Beat is terminated, you can check the envelopes
    assert len(envelopes) >= 0


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
@pytest.mark.forked
def test_beat_task_crons_success(celery_init, capture_envelopes):
    app = celery_init(
        monitor_beat_tasks=True,
    )
    envelopes = capture_envelopes()

    @app.task
    def test_task():
        logger.info("Running test_task")

    app.add_periodic_task(60.0, test_task.s(), name="success_from_beat")

    with start_worker(app, perform_ping_check=False):
        run_beat(app)

    assert len(envelopes) == 2
    (envelop_in_progress, envelope_ok) = envelopes

    assert envelop_in_progress.items[0].headers["type"] == "check_in"
    check_in = envelop_in_progress.items[0].payload.json
    assert check_in["type"] == "check_in"
    assert check_in["monitor_slug"] == "success_from_beat"
    assert check_in["status"] == "in_progress"

    assert envelope_ok.items[0].headers["type"] == "check_in"
    check_in = envelope_ok.items[0].payload.json
    assert check_in["type"] == "check_in"
    assert check_in["monitor_slug"] == "success_from_beat"
    assert check_in["status"] == "ok"


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
@pytest.mark.forked
def test_beat_task_crons_error(celery_init, capture_envelopes):
    app = celery_init(
        monitor_beat_tasks=True,
    )
    envelopes = capture_envelopes()

    @app.task
    def test_task():
        logger.info("Running test_task")
        1 / 0

    app.add_periodic_task(60.0, test_task.s(), name="failure_from_beat")

    with start_worker(app, perform_ping_check=False):
        run_beat(app)

    envelop_in_progress = envelopes[0]
    envelope_error = envelopes[-1]

    check_in = envelop_in_progress.items[0].payload.json
    assert check_in["type"] == "check_in"
    assert check_in["monitor_slug"] == "failure_from_beat"
    assert check_in["status"] == "in_progress"

    check_in = envelope_error.items[0].payload.json
    assert check_in["type"] == "check_in"
    assert check_in["monitor_slug"] == "failure_from_beat"
    assert check_in["status"] == "error"



import asyncio
import json
import threading

from django.contrib.auth import login
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.dispatch import Signal
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseServerError
from django.shortcuts import render
from django.template import Context, Template
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import ListView

from tests.integrations.django.myapp.signals import (
    myapp_custom_signal,
    myapp_custom_signal_silenced,
)

try:
    from rest_framework.decorators import api_view
    from rest_framework.response import Response

    @api_view(["POST"])
    def rest_framework_exc(request):
        1 / 0

    @api_view(["POST"])
    def rest_framework_read_body_and_exc(request):
        request.data
        1 / 0

    @api_view(["GET"])
    def rest_hello(request):
        return HttpResponse("ok")

    @api_view(["GET"])
    def rest_permission_denied_exc(request):
        raise PermissionDenied("bye")

    @api_view(["GET"])
    def rest_json_response(request):
        return Response(dict(ok=True))

except ImportError:
    pass


import sentry_sdk
from sentry_sdk import capture_message


@csrf_exempt
def view_exc(request):
    1 / 0


@csrf_exempt
def view_exc_with_msg(request):
    capture_message("oops")
    1 / 0


@cache_page(60)
def cached_view(request):
    return HttpResponse("ok")


def not_cached_view(request):
    return HttpResponse("ok")


def view_with_cached_template_fragment(request):
    template = Template(
        """{% load cache %}
        Not cached content goes here.
        {% cache 500 some_identifier %}
            And here some cached content.
        {% endcache %}
        """
    )
    rendered = template.render(Context({}))
    return HttpResponse(rendered)


# This is a "class based view" as previously found in the sentry codebase. The
# interesting property of this one is that csrf_exempt, as a class attribute,
# is not in __dict__, so regular use of functools.wraps will not forward the
# attribute.
class SentryClassBasedView:
    csrf_exempt = True

    def __call__(self, request):
        return HttpResponse("ok")


class SentryClassBasedViewWithCsrf:
    def __call__(self, request):
        return HttpResponse("ok")


@csrf_exempt
def read_body_and_view_exc(request):
    request.read()
    1 / 0


@csrf_exempt
def message(request):
    sentry_sdk.capture_message("hi")
    return HttpResponse("ok")


@csrf_exempt
def nomessage(request):
    return HttpResponse("ok")


@csrf_exempt
def view_with_signal(request):
    custom_signal = Signal()
    custom_signal.send(sender="hello")
    return HttpResponse("ok")


@csrf_exempt
def mylogin(request):
    user = User.objects.create_user("john", "lennon@thebeatles.com", "johnpassword")
    user.backend = "django.contrib.auth.backends.ModelBackend"
    login(request, user)
    return HttpResponse("ok")


@csrf_exempt
def handler500(request):
    return HttpResponseServerError("Sentry error.")


class ClassBasedView(ListView):
    model = None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def head(self, *args, **kwargs):
        sentry_sdk.capture_message("hi")
        return HttpResponse("")

    def post(self, *args, **kwargs):
        return HttpResponse("ok")


@csrf_exempt
def post_echo(request):
    sentry_sdk.capture_message("hi")
    return HttpResponse(request.body)


@csrf_exempt
def handler404(*args, **kwargs):
    sentry_sdk.capture_message("not found", level="error")
    return HttpResponseNotFound("404")


@csrf_exempt
def template_exc(request, *args, **kwargs):
    return render(request, "error.html")


@csrf_exempt
def template_test(request, *args, **kwargs):
    return render(request, "user_name.html", {"user_age": 20})


@csrf_exempt
def custom_ok(request, *args, **kwargs):
    return HttpResponse("custom ok")


@csrf_exempt
def custom_exc(request, *args, **kwargs):
    1 / 0


@csrf_exempt
def template_test2(request, *args, **kwargs):
    return TemplateResponse(
        request, ("user_name.html", "another_template.html"), {"user_age": 25}
    )


@csrf_exempt
def template_test3(request, *args, **kwargs):
    traceparent = sentry_sdk.get_current_scope().get_traceparent()
    if traceparent is None:
        traceparent = sentry_sdk.get_isolation_scope().get_traceparent()

    baggage = sentry_sdk.get_current_scope().get_baggage()
    if baggage is None:
        baggage = sentry_sdk.get_isolation_scope().get_baggage()

    capture_message(traceparent + "\n" + baggage.serialize())
    return render(request, "trace_meta.html", {})


@csrf_exempt
def template_test4(request, *args, **kwargs):
    User.objects.create_user("john", "lennon@thebeatles.com", "johnpassword")
    my_queryset = User.objects.all()  # noqa

    template_context = {
        "user_age": 25,
        "complex_context": my_queryset,
        "complex_list": [1, 2, 3, my_queryset],
        "complex_dict": {
            "a": 1,
            "d": my_queryset,
        },
        "none_context": None,
    }

    return TemplateResponse(
        request,
        "user_name.html",
        template_context,
    )


@csrf_exempt
def postgres_select(request, *args, **kwargs):
    from django.db import connections

    cursor = connections["postgres"].cursor()
    cursor.execute("SELECT 1;")
    return HttpResponse("ok")


@csrf_exempt
def postgres_select_orm(request, *args, **kwargs):
    user = User.objects.using("postgres").all().first()
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def postgres_insert_orm_no_autocommit(request, *args, **kwargs):
    transaction.set_autocommit(False, using="postgres")
    try:
        user = User.objects.db_manager("postgres").create_user(
            username="user1",
        )
        transaction.commit(using="postgres")
    except Exception:
        transaction.rollback(using="postgres")
        transaction.set_autocommit(True, using="postgres")
        raise

    transaction.set_autocommit(True, using="postgres")
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def postgres_insert_orm_no_autocommit_rollback(request, *args, **kwargs):
    transaction.set_autocommit(False, using="postgres")
    try:
        user = User.objects.db_manager("postgres").create_user(
            username="user1",
        )
        transaction.rollback(using="postgres")
    except Exception:
        transaction.rollback(using="postgres")
        transaction.set_autocommit(True, using="postgres")
        raise

    transaction.set_autocommit(True, using="postgres")
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def postgres_insert_orm_atomic(request, *args, **kwargs):
    with transaction.atomic(using="postgres"):
        user = User.objects.db_manager("postgres").create_user(
            username="user1",
        )
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def postgres_insert_orm_atomic_rollback(request, *args, **kwargs):
    with transaction.atomic(using="postgres"):
        user = User.objects.db_manager("postgres").create_user(
            username="user1",
        )
        transaction.set_rollback(True, using="postgres")
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def postgres_insert_orm_atomic_exception(request, *args, **kwargs):
    try:
        with transaction.atomic(using="postgres"):
            user = User.objects.db_manager("postgres").create_user(
                username="user1",
            )
            transaction.set_rollback(True, using="postgres")
            1 / 0
    except ZeroDivisionError:
        pass
    return HttpResponse("ok {}".format(user))


@csrf_exempt
def permission_denied_exc(*args, **kwargs):
    raise PermissionDenied("bye")


def csrf_hello_not_exempt(*args, **kwargs):
    return HttpResponse("ok")


def thread_ids_sync(*args, **kwargs):
    response = json.dumps(
        {
            "main": threading.main_thread().ident,
            "active": threading.current_thread().ident,
        }
    )
    return HttpResponse(response)


async def async_message(request):
    sentry_sdk.capture_message("hi")
    return HttpResponse("ok")


async def my_async_view(request):
    await asyncio.sleep(1)
    return HttpResponse("Hello World")


async def simple_async_view(request):
    return HttpResponse("Simple Hello World")


async def thread_ids_async(request):
    response = json.dumps(
        {
            "main": threading.main_thread().ident,
            "active": threading.current_thread().ident,
        }
    )
    return HttpResponse(response)


async def post_echo_async(request):
    sentry_sdk.capture_message("hi")
    return HttpResponse(request.body)


post_echo_async.csrf_exempt = True


@csrf_exempt
def send_myapp_custom_signal(request):
    myapp_custom_signal.send(sender="hello")
    myapp_custom_signal_silenced.send(sender="hello")
    return HttpResponse("ok")



#!/usr/bin/env python

"""
Sentry-Python - Sentry SDK for Python
=====================================

**Sentry-Python is an SDK for Sentry.** Check out `GitHub
<https://github.com/getsentry/sentry-python>`_ to find out more.
"""

import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def get_file_text(file_name):
    with open(os.path.join(here, file_name)) as in_file:
        return in_file.read()


setup(
    name="sentry-sdk",
    version="2.66.1",
    author="Sentry Team and Contributors",
    author_email="hello@sentry.io",
    url="https://github.com/getsentry/sentry-python",
    project_urls={
        "Documentation": "https://docs.sentry.io/platforms/python/",
        "Changelog": "https://github.com/getsentry/sentry-python/blob/master/CHANGELOG.md",
    },
    description="Python client for Sentry (https://sentry.io)",
    long_description=get_file_text("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", "tests.*")),
    # PEP 561
    package_data={"sentry_sdk": ["py.typed"]},
    zip_safe=False,
    license_expression="MIT",
    python_requires=">=3.6",
    install_requires=[
        "urllib3>=1.26.11",
        "certifi",
    ],
    extras_require={
        "aiohttp": ["aiohttp>=3.5"],
        "anthropic": ["anthropic>=0.16"],
        "arq": ["arq>=0.23"],
        "asyncpg": ["asyncpg>=0.23"],
        "beam": ["apache-beam>=2.12"],
        "bottle": ["bottle>=0.12.13"],
        "celery": ["celery>=3"],
        "celery-redbeat": ["celery-redbeat>=2"],
        "chalice": ["chalice>=1.16.0"],
        "clickhouse-driver": ["clickhouse-driver>=0.2.0"],
        "django": ["django>=1.8"],
        "falcon": ["falcon>=1.4"],
        "fastapi": ["fastapi>=0.79.0"],
        "flask": ["flask>=0.11", "blinker>=1.1", "markupsafe"],
        "grpcio": ["grpcio>=1.21.1", "protobuf>=3.8.0"],
        "http2": ["httpcore[http2]==1.*"],
        "asyncio": ["httpcore[asyncio]==1.*"],
        "httpx": ["httpx>=0.16.0"],
        "huey": ["huey>=2"],
        "huggingface_hub": ["huggingface_hub>=0.22"],
        "langchain": ["langchain>=0.0.210"],
        "langgraph": ["langgraph>=0.6.6"],
        "launchdarkly": ["launchdarkly-server-sdk>=9.8.0"],
        "litellm": ["litellm>=1.77.5,!=1.82.7,!=1.82.8"],
        "litestar": ["litestar>=2.0.0"],
        "loguru": ["loguru>=0.5"],
        "mcp": ["mcp>=1.15.0"],
        "openai": ["openai>=1.0.0", "tiktoken>=0.3.0"],
        "openfeature": ["openfeature-sdk>=0.7.1"],
        "opentelemetry": ["opentelemetry-distro>=0.35b0"],
        "opentelemetry-experimental": ["opentelemetry-distro"],
        "opentelemetry-otlp": ["opentelemetry-distro[otlp]>=0.35b0"],
        "pure-eval": ["pure_eval", "executing", "asttokens"],
        "pydantic_ai": ["pydantic-ai>=1.0.0"],
        "pymongo": ["pymongo>=3.1"],
        "pyspark": ["pyspark>=2.4.4"],
        "quart": ["quart>=0.16.1", "blinker>=1.1"],
        "rq": ["rq>=0.6"],
        "sanic": ["sanic>=0.8"],
        "sqlalchemy": ["sqlalchemy>=1.2"],
        "starlette": ["starlette>=0.19.1"],
        "starlite": ["starlite>=1.48"],
        "statsig": ["statsig>=0.55.3"],
        "tornado": ["tornado>=6"],
        "unleash": ["UnleashClient>=6.0.1"],
        "google-genai": ["google-genai>=1.29.0"],
    },
    entry_points={
        "opentelemetry_propagator": [
            "sentry=sentry_sdk.integrations.opentelemetry:SentryPropagator"
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import asyncio
import base64
import inspect
import json
import os
import sys
from unittest import mock

import django
import pytest
from channels.testing import HttpCommunicator

import sentry_sdk
from sentry_sdk import capture_message
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.django import DjangoIntegration
from sentry_sdk.integrations.django.asgi import _asgi_middleware_mixin_factory
from tests.integrations.django.myapp.asgi import channels_application

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


APPS = [channels_application]
if django.VERSION >= (3, 0):
    from tests.integrations.django.myapp.asgi import asgi_application

    APPS += [asgi_application]


@pytest.fixture
def make_asgi_application():
    """Build a fresh ASGI application. Call AFTER sentry_init so middleware
    instrumentation is installed before Django builds the middleware chain."""
    from django.core.asgi import get_asgi_application

    return get_asgi_application


@pytest.mark.parametrize("application", APPS)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 0), reason="Django ASGI support shipped in 3.0"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_basic(
    sentry_init,
    capture_events,
    capture_items,
    application,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    import channels  # type: ignore[import-not-found]

    if span_streaming:
        items = capture_items("event")

        if (
            sys.version_info < (3, 9)
            and channels.__version__ < "4.0.0"
            and django.VERSION >= (3, 0)
            and django.VERSION < (4, 0)
        ):
            # We emit a UserWarning for channels 2.x and 3.x on Python 3.8 and older
            # because the async support was not really good back then and there is a known issue.
            # See the TreadingIntegration for details.
            with pytest.warns(UserWarning):
                comm = HttpCommunicator(application, "GET", "/view-exc?test=query")
                response = await comm.get_response()
                await comm.wait()
        else:
            comm = HttpCommunicator(application, "GET", "/view-exc?test=query")
            response = await comm.get_response()
            await comm.wait()

        assert response["status"] == 500

        (event,) = (item.payload for item in items if item.type == "event")

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"

        # Test that the ASGI middleware got set up correctly. Right now this needs
        # to be installed manually (see myapp/asgi.py)
        assert event["transaction"] == "/view-exc"
        assert event["request"] == {
            "cookies": {},
            "headers": {},
            "method": "GET",
            "query_string": "test=query",
            "url": "/view-exc",
        }

        capture_message("hi")
        event = items[-1].payload
    else:
        events = capture_events()

        if (
            sys.version_info < (3, 9)
            and channels.__version__ < "4.0.0"
            and django.VERSION >= (3, 0)
            and django.VERSION < (4, 0)
        ):
            # We emit a UserWarning for channels 2.x and 3.x on Python 3.8 and older
            # because the async support was not really good back then and there is a known issue.
            # See the TreadingIntegration for details.
            with pytest.warns(UserWarning):
                comm = HttpCommunicator(application, "GET", "/view-exc?test=query")
                response = await comm.get_response()
                await comm.wait()
        else:
            comm = HttpCommunicator(application, "GET", "/view-exc?test=query")
            response = await comm.get_response()
            await comm.wait()

        assert response["status"] == 500

        (event,) = events

        (exception,) = event["exception"]["values"]
        assert exception["type"] == "ZeroDivisionError"

        # Test that the ASGI middleware got set up correctly. Right now this needs
        # to be installed manually (see myapp/asgi.py)
        assert event["transaction"] == "/view-exc"
        assert event["request"] == {
            "cookies": {},
            "headers": {},
            "method": "GET",
            "query_string": "test=query",
            "url": "/view-exc",
        }

        capture_message("hi")
        event = events[-1]

    assert "request" not in event


@pytest.mark.parametrize("application", APPS)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_async_views(
    sentry_init,
    capture_events,
    capture_items,
    application,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(application, "GET", "/async_message")
    if span_streaming:
        items = capture_items("event")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200

        (event,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200

        (event,) = events

    assert event["transaction"] == "/async_message"
    assert event["request"] == {
        "cookies": {},
        "headers": {},
        "method": "GET",
        "query_string": None,
        "url": "/async_message",
    }


@pytest.mark.parametrize("application", APPS)
@pytest.mark.parametrize("endpoint", ["/sync/thread_ids", "/async/thread_ids"])
@pytest.mark.parametrize("middleware_spans", [False, True])
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_active_thread_id(
    sentry_init,
    capture_envelopes,
    capture_items,
    teardown_profiling,
    endpoint,
    application,
    middleware_spans,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration(middleware_spans=middleware_spans)],
        traces_sample_rate=1.0,
        profiles_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(application, "GET", endpoint)

    if span_streaming:
        with mock.patch(
            "sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0
        ):
            items = capture_items("span")
            response = await comm.get_response()
            await comm.wait()

            assert response["status"] == 200, response["body"]

        data = json.loads(response["body"])

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        for span in spans:
            if span["is_segment"] is False:
                continue
            assert str(data["active"]) == span["attributes"]["thread.id"]
    else:
        with mock.patch(
            "sentry_sdk.profiler.transaction_profiler.PROFILE_MINIMUM_SAMPLES", 0
        ):
            envelopes = capture_envelopes()
            response = await comm.get_response()
            await comm.wait()

            assert response["status"] == 200, response["body"]

        assert len(envelopes) == 1

        profiles = [item for item in envelopes[0].items if item.type == "profile"]
        assert len(profiles) == 1

        data = json.loads(response["body"])

        for item in profiles:
            transactions = item.payload.json["transactions"]
            assert len(transactions) == 1
            assert str(data["active"]) == transactions[0]["active_thread_id"]

        transactions = [
            item for item in envelopes[0].items if item.type == "transaction"
        ]
        assert len(transactions) == 1

        for item in transactions:
            transaction = item.payload.json
            trace_context = transaction["contexts"]["trace"]
            assert str(data["active"]) == trace_context["data"]["thread.id"]


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
async def test_async_views_concurrent_execution(
    sentry_init, settings, make_asgi_application
):
    import time

    settings.MIDDLEWARE = []
    sentry_init(
        integrations=[DjangoIntegration()],
        send_default_pii=True,
    )

    application = make_asgi_application()

    comm = HttpCommunicator(application, "GET", "/my_async_view")  # sleeps for 1 second
    comm2 = HttpCommunicator(
        application, "GET", "/my_async_view"
    )  # sleeps for 1 second

    start = time.monotonic()
    resp1, resp2 = await asyncio.gather(
        comm.get_response(timeout=5),
        comm2.get_response(timeout=5),
    )
    await asyncio.gather(comm.wait(), comm2.wait())
    end = time.monotonic()

    assert resp1["status"] == 200
    assert resp2["status"] == 200

    assert (
        end - start < 2
    )  # it takes less than 2 seconds so it was ececuting concurrently


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
async def test_async_middleware_that_is_function_concurrent_execution(
    sentry_init, settings, make_asgi_application
):
    import time

    settings.MIDDLEWARE = [
        "tests.integrations.django.myapp.middleware.simple_middleware"
    ]
    sentry_init(
        integrations=[DjangoIntegration()],
        send_default_pii=True,
    )

    application = make_asgi_application()

    comm = HttpCommunicator(application, "GET", "/my_async_view")  # sleeps for 1 second
    comm2 = HttpCommunicator(
        application, "GET", "/my_async_view"
    )  # sleeps for 1 second

    start = time.monotonic()
    resp1, resp2 = await asyncio.gather(
        comm.get_response(timeout=5),
        comm2.get_response(timeout=5),
    )
    await asyncio.gather(comm.wait(), comm2.wait())
    end = time.monotonic()

    assert resp1["status"] == 200
    assert resp2["status"] == 200

    assert (
        end - start < 2
    )  # it takes less than 2 seconds so it was ececuting concurrently


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_async_middleware_spans(
    sentry_init,
    render_span_tree,
    capture_events,
    capture_items,
    settings,
    span_streaming,
    make_asgi_application,
):
    settings.MIDDLEWARE = [
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "tests.integrations.django.myapp.settings.TestMiddleware",
    ]
    sentry_init(
        integrations=[DjangoIntegration(middleware_spans=True)],
        traces_sample_rate=1.0,
        _experiments={
            "record_sql_params": True,
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    application = make_asgi_application()

    comm = HttpCommunicator(application, "GET", "/simple_async_view")
    if span_streaming:
        items = capture_items("span")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # Filter out signal-receiver spans — their ordering depends on Django
        # module import order and is not what this middleware test verifies.
        spans = [s for s in spans if s["attributes"].get("sentry.op") != "event.django"]

        assert (
            render_span_tree(spans)
            == """\
- sentry.op="http.server": name="/simple_async_view"
  - sentry.op="middleware.django": name="django.contrib.sessions.middleware.SessionMiddleware.__acall__"
    - sentry.op="middleware.django": name="django.contrib.auth.middleware.AuthenticationMiddleware.__acall__"
      - sentry.op="middleware.django": name="django.middleware.csrf.CsrfViewMiddleware.__acall__"
        - sentry.op="middleware.django": name="tests.integrations.django.myapp.settings.TestMiddleware.__acall__"
          - sentry.op="middleware.django": name="django.middleware.csrf.CsrfViewMiddleware.process_view"
          - sentry.op="view.render": name="simple_async_view\""""
        )
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200

        (transaction,) = events

        assert transaction["type"] == "transaction"

        # Filter out signal-receiver spans — their ordering depends on Django
        # module import order and is not what this middleware test verifies.
        spans = [s for s in transaction["spans"] if s.get("op") != "event.django"]

        assert (
            render_span_tree(spans, transaction["contexts"]["trace"])
            == """\
- op="http.server": description=null
  - op="middleware.django": description="django.contrib.sessions.middleware.SessionMiddleware.__acall__"
    - op="middleware.django": description="django.contrib.auth.middleware.AuthenticationMiddleware.__acall__"
      - op="middleware.django": description="django.middleware.csrf.CsrfViewMiddleware.__acall__"
        - op="middleware.django": description="tests.integrations.django.myapp.settings.TestMiddleware.__acall__"
          - op="middleware.django": description="django.middleware.csrf.CsrfViewMiddleware.process_view"
          - op="view.render": description="simple_async_view\""""
        )


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_has_trace_if_performance_enabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(asgi_application, "GET", "/view-exc-with-msg")
    if span_streaming:
        items = capture_items("event", "span")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (
            msg_event,
            error_event,
        ) = (item.payload for item in items if item.type == "event")

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        assert spans[6]["is_segment"] is True

        assert (
            msg_event["contexts"]["trace"]["trace_id"]
            == error_event["contexts"]["trace"]["trace_id"]
            == spans[6]["trace_id"]
        )
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (msg_event, error_event, transaction_event) = events

        assert (
            msg_event["contexts"]["trace"]["trace_id"]
            == error_event["contexts"]["trace"]["trace_id"]
            == transaction_event["contexts"]["trace"]["trace_id"]
        )


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_has_trace_if_performance_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(asgi_application, "GET", "/view-exc-with-msg")
    if span_streaming:
        items = capture_items("event")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (
            msg_event,
            error_event,
        ) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (msg_event, error_event) = events

    assert msg_event["contexts"]["trace"]
    assert "trace_id" in msg_event["contexts"]["trace"]

    assert error_event["contexts"]["trace"]
    assert "trace_id" in error_event["contexts"]["trace"]
    assert (
        msg_event["contexts"]["trace"]["trace_id"]
        == error_event["contexts"]["trace"]["trace_id"]
    )


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_trace_from_headers_if_performance_enabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    trace_id = "582b43a4192642f0b136d5159a501701"
    sentry_trace_header = "{}-{}-{}".format(trace_id, "6e8f22c393e68f19", 1)

    comm = HttpCommunicator(
        asgi_application,
        "GET",
        "/view-exc-with-msg",
        headers=[(b"sentry-trace", sentry_trace_header.encode())],
    )

    if span_streaming:
        items = capture_items("event", "span")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (
            msg_event,
            error_event,
        ) = (item.payload for item in items if item.type == "event")

        assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
        assert error_event["contexts"]["trace"]["trace_id"] == trace_id

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]
        assert spans[6]["is_segment"] is True
        assert spans[6]["trace_id"] == trace_id
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (msg_event, error_event, transaction_event) = events

        assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
        assert error_event["contexts"]["trace"]["trace_id"] == trace_id
        assert transaction_event["contexts"]["trace"]["trace_id"] == trace_id


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_trace_from_headers_if_performance_disabled(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    trace_id = "582b43a4192642f0b136d5159a501701"
    sentry_trace_header = "{}-{}-{}".format(trace_id, "6e8f22c393e68f19", 1)

    comm = HttpCommunicator(
        asgi_application,
        "GET",
        "/view-exc-with-msg",
        headers=[(b"sentry-trace", sentry_trace_header.encode())],
    )

    if span_streaming:
        items = capture_items("event")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (msg_event, error_event) = (
            item.payload for item in items if item.type == "event"
        )
    else:
        events = capture_events()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 500

        (msg_event, error_event) = events

    assert msg_event["contexts"]["trace"]["trace_id"] == trace_id
    assert error_event["contexts"]["trace"]["trace_id"] == trace_id


PICTURE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "image.png")
BODY_FORM = """--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="username"\r\n\r\nJane\r\n--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="password"\r\n\r\nhello123\r\n--fd721ef49ea403a6\r\nContent-Disposition: form-data; name="photo"; filename="image.png"\r\nContent-Type: image/png\r\nContent-Transfer-Encoding: base64\r\n\r\n{{image_data}}\r\n--fd721ef49ea403a6--\r\n""".replace(
    "{{image_data}}", base64.b64encode(open(PICTURE, "rb").read()).decode("utf-8")
).encode("utf-8")
BODY_FORM_CONTENT_LENGTH = str(len(BODY_FORM)).encode("utf-8")


@pytest.mark.parametrize("application", APPS)
@pytest.mark.parametrize(
    "send_default_pii,method,headers,url_name,body,expected_data",
    [
        (
            True,
            "POST",
            [(b"content-type", b"text/plain")],
            "post_echo_async",
            b"",
            None,
        ),
        (
            True,
            "POST",
            [(b"content-type", b"text/plain")],
            "post_echo_async",
            b"some raw text body",
            "",
        ),
        (
            True,
            "POST",
            [(b"content-type", b"application/json")],
            "post_echo_async",
            b'{"username":"xyz","password":"xyz"}',
            {"username": "xyz", "password": "[Filtered]"},
        ),
        (
            True,
            "POST",
            [(b"content-type", b"application/xml")],
            "post_echo_async",
            b'<?xml version="1.0" encoding="UTF-8"?><root></root>',
            "",
        ),
        (
            True,
            "POST",
            [
                (b"content-type", b"multipart/form-data; boundary=fd721ef49ea403a6"),
                (b"content-length", BODY_FORM_CONTENT_LENGTH),
            ],
            "post_echo_async",
            BODY_FORM,
            {"password": "[Filtered]", "photo": "", "username": "Jane"},
        ),
        (
            False,
            "POST",
            [(b"content-type", b"text/plain")],
            "post_echo_async",
            b"",
            None,
        ),
        (
            False,
            "POST",
            [(b"content-type", b"text/plain")],
            "post_echo_async",
            b"some raw text body",
            "",
        ),
        (
            False,
            "POST",
            [(b"content-type", b"application/json")],
            "post_echo_async",
            b'{"username":"xyz","password":"xyz"}',
            {"username": "xyz", "password": "[Filtered]"},
        ),
        (
            False,
            "POST",
            [(b"content-type", b"application/xml")],
            "post_echo_async",
            b'<?xml version="1.0" encoding="UTF-8"?><root></root>',
            "",
        ),
        (
            False,
            "POST",
            [
                (b"content-type", b"multipart/form-data; boundary=fd721ef49ea403a6"),
                (b"content-length", BODY_FORM_CONTENT_LENGTH),
            ],
            "post_echo_async",
            BODY_FORM,
            {"password": "[Filtered]", "photo": "", "username": "Jane"},
        ),
    ],
)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_asgi_request_body(
    sentry_init,
    capture_envelopes,
    capture_items,
    application,
    send_default_pii,
    method,
    headers,
    url_name,
    body,
    expected_data,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(
        application,
        method=method,
        headers=headers,
        path=reverse(url_name),
        body=body,
    )

    if span_streaming:
        items = capture_items("event")

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200
        assert response["body"] == body

        sentry_sdk.flush()
        (event,) = (item.payload for item in items if item.type == "event")
    else:
        envelopes = capture_envelopes()

        response = await comm.get_response()
        await comm.wait()

        assert response["status"] == 200
        assert response["body"] == body

        (envelope,) = envelopes
        event = envelope.get_event()

    if expected_data is not None:
        assert event["request"]["data"] == expected_data
    else:
        assert "data" not in event["request"]


@pytest.mark.asyncio
@pytest.mark.skipif(
    sys.version_info >= (3, 12),
    reason=(
        "asyncio.iscoroutinefunction has been replaced in 3.12 by inspect.iscoroutinefunction"
    ),
)
@pytest.mark.skipif(
    django.VERSION < (3, 0), reason="Django ASGI support shipped in 3.0"
)
async def test_asgi_mixin_iscoroutinefunction_before_3_12():
    sentry_asgi_mixin = _asgi_middleware_mixin_factory(lambda: None)

    async def get_response(): ...

    instance = sentry_asgi_mixin(get_response)
    assert asyncio.iscoroutinefunction(instance)


@pytest.mark.skipif(
    sys.version_info >= (3, 12),
    reason=(
        "asyncio.iscoroutinefunction has been replaced in 3.12 by inspect.iscoroutinefunction"
    ),
)
def test_asgi_mixin_iscoroutinefunction_when_not_async_before_3_12():
    sentry_asgi_mixin = _asgi_middleware_mixin_factory(lambda: None)

    def get_response(): ...

    instance = sentry_asgi_mixin(get_response)
    assert not asyncio.iscoroutinefunction(instance)


@pytest.mark.asyncio
@pytest.mark.skipif(
    sys.version_info < (3, 12),
    reason=(
        "asyncio.iscoroutinefunction has been replaced in 3.12 by inspect.iscoroutinefunction"
    ),
)
async def test_asgi_mixin_iscoroutinefunction_after_3_12():
    sentry_asgi_mixin = _asgi_middleware_mixin_factory(lambda: None)

    async def get_response(): ...

    instance = sentry_asgi_mixin(get_response)
    assert inspect.iscoroutinefunction(instance)


@pytest.mark.skipif(
    sys.version_info < (3, 12),
    reason=(
        "asyncio.iscoroutinefunction has been replaced in 3.12 by inspect.iscoroutinefunction"
    ),
)
def test_asgi_mixin_iscoroutinefunction_when_not_async_after_3_12():
    sentry_asgi_mixin = _asgi_middleware_mixin_factory(lambda: None)

    def get_response(): ...

    instance = sentry_asgi_mixin(get_response)
    assert not inspect.iscoroutinefunction(instance)


@pytest.mark.parametrize("application", APPS)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="async views have been introduced in Django 3.1"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_async_view(
    sentry_init,
    capture_events,
    capture_items,
    application,
    span_streaming,
):
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    comm = HttpCommunicator(application, "GET", "/simple_async_view")
    if span_streaming:
        items = capture_items("span")

        await comm.get_response()
        await comm.wait()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[5]["name"] == "/simple_async_view"
    else:
        events = capture_events()

        await comm.get_response()
        await comm.wait()

        (event,) = events
        assert event["type"] == "transaction"
        assert event["transaction"] == "/simple_async_view"


@pytest.mark.parametrize("application", APPS)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 0), reason="Django ASGI support shipped in 3.0"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_transaction_http_method_default(
    sentry_init,
    capture_events,
    capture_items,
    application,
    span_streaming,
):
    """
    By default OPTIONS and HEAD requests do not create a transaction.
    """
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        comm = HttpCommunicator(application, "GET", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "OPTIONS", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "HEAD", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[5]["attributes"]["http.request.method"] == "GET"
    else:
        events = capture_events()

        comm = HttpCommunicator(application, "GET", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "OPTIONS", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "HEAD", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        (event,) = events

        assert len(events) == 1
        assert event["request"]["method"] == "GET"


@pytest.mark.parametrize("application", APPS)
@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 0), reason="Django ASGI support shipped in 3.0"
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_transaction_http_method_custom(
    sentry_init,
    capture_events,
    capture_items,
    application,
    span_streaming,
):
    sentry_init(
        integrations=[
            DjangoIntegration(
                http_methods_to_capture=(
                    "OPTIONS",
                    "head",
                ),  # capitalization does not matter
            )
        ],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        comm = HttpCommunicator(application, "GET", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "OPTIONS", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "HEAD", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[5]["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "OPTIONS"
        assert spans[11]["attributes"][SPANDATA.HTTP_REQUEST_METHOD] == "HEAD"
    else:
        events = capture_events()

        comm = HttpCommunicator(application, "GET", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "OPTIONS", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        comm = HttpCommunicator(application, "HEAD", "/simple_async_view")
        await comm.get_response()
        await comm.wait()

        assert len(events) == 2

        (event1, event2) = events
        assert event1["request"]["method"] == "OPTIONS"
        assert event2["request"]["method"] == "HEAD"


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1),
    reason="async views/middleware introduced in Django 3.1",
)
async def test_async_middleware_process_view_is_awaited(
    sentry_init, settings, make_asgi_application
):
    """Regression test for async ``process_view`` being coerced to sync."""
    sentry_init(integrations=[DjangoIntegration()])

    settings.MIDDLEWARE = [
        "tests.integrations.django.myapp.middleware.AsyncProcessViewMiddleware"
    ]
    application = make_asgi_application()

    comm = HttpCommunicator(application, "GET", "/simple_async_view")
    response = await comm.get_response()
    await comm.wait()

    assert response["status"] == 200


@pytest.mark.asyncio
@pytest.mark.skipif(
    django.VERSION < (3, 1),
    reason="async views/middleware introduced in Django 3.1",
)
async def test_async_middleware_process_exception_is_awaited(
    sentry_init, settings, make_asgi_application
):
    """Regression test for async ``process_exception`` being coerced to sync."""
    sentry_init(integrations=[DjangoIntegration()])

    settings.MIDDLEWARE = [
        "tests.integrations.django.myapp.middleware.AsyncProcessExceptionMiddleware"
    ]
    application = make_asgi_application()

    comm = HttpCommunicator(application, "GET", "/view-exc")
    response = await comm.get_response()
    await comm.wait()

    assert response["status"] == 200
    assert response["body"] == b"handled by async process_exception"
